{% if errors %}
  <tr>
    <td colspan="2">
      {{ errors }}
      {% if not fields %}
        {% for field in hidden_fields %}{{ field }}{% endfor %}
      {% endif %}
    </td>
  </tr>
{% endif %}
{% for field, errors in fields %}
  <tr{% set classes = field.css_classes() %}{% if classes %} class="{{ classes }}"{% endif %}>
    <th>{% if field.label %}{{ field.label_tag() }}{% endif %}</th>
    <td>
      {{ errors }}
      {{ field }}
      {% if field.help_text %}
        <br>
        <span class="helptext"{% if field.auto_id %} id="{{ field.auto_id }}_helptext"{% endif %}>{{ field.help_text|safe }}</span>
      {% endif %}
      {% if loop.last %}
        {% for field in hidden_fields %}{{ field }}{% endfor %}
      {% endif %}
    </td>
  </tr>
{% endfor %}
{% if not fields and not errors %}
  {% for field in hidden_fields %}{{ field }}{% endfor %}
{% endif %}
