Recursion!
{% for comment in comments %}
    {{ comment.comment }}
    {% if comment.children %}
        {% include "./recursive_relative_include.html" with comments=comment.children %}
    {% endif %}
{% endfor %}
